# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: remote.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cremote.proto\x12\x06remote\"-\n\x06Sample\x12\r\n\x05value\x18\x01 \x01(\x01\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\"(\n\tLabelPair\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"P\n\nTimeSeries\x12!\n\x06labels\x18\x01 \x03(\x0b\x32\x11.remote.LabelPair\x12\x1f\n\x07samples\x18\x02 \x03(\x0b\x32\x0e.remote.Sample\"6\n\x0cWriteRequest\x12&\n\ntimeseries\x18\x01 \x03(\x0b\x32\x12.remote.TimeSeries\"-\n\x0bReadRequest\x12\x1e\n\x07queries\x18\x01 \x03(\x0b\x32\r.remote.Query\"4\n\x0cReadResponse\x12$\n\x07results\x18\x01 \x03(\x0b\x32\x13.remote.QueryResult\"e\n\x05Query\x12\x1a\n\x12start_timestamp_ms\x18\x01 \x01(\x03\x12\x18\n\x10\x65nd_timestamp_ms\x18\x02 \x01(\x03\x12&\n\x08matchers\x18\x03 \x03(\x0b\x32\x14.remote.LabelMatcher\"L\n\x0cLabelMatcher\x12\x1f\n\x04type\x18\x01 \x01(\x0e\x32\x11.remote.MatchType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"5\n\x0bQueryResult\x12&\n\ntimeseries\x18\x01 \x03(\x0b\x32\x12.remote.TimeSeries*J\n\tMatchType\x12\t\n\x05\x45QUAL\x10\x00\x12\r\n\tNOT_EQUAL\x10\x01\x12\x0f\n\x0bREGEX_MATCH\x10\x02\x12\x12\n\x0eREGEX_NO_MATCH\x10\x03\x62\x06proto3')

_MATCHTYPE = DESCRIPTOR.enum_types_by_name['MatchType']
MatchType = enum_type_wrapper.EnumTypeWrapper(_MATCHTYPE)
EQUAL = 0
NOT_EQUAL = 1
REGEX_MATCH = 2
REGEX_NO_MATCH = 3


_SAMPLE = DESCRIPTOR.message_types_by_name['Sample']
_LABELPAIR = DESCRIPTOR.message_types_by_name['LabelPair']
_TIMESERIES = DESCRIPTOR.message_types_by_name['TimeSeries']
_WRITEREQUEST = DESCRIPTOR.message_types_by_name['WriteRequest']
_READREQUEST = DESCRIPTOR.message_types_by_name['ReadRequest']
_READRESPONSE = DESCRIPTOR.message_types_by_name['ReadResponse']
_QUERY = DESCRIPTOR.message_types_by_name['Query']
_LABELMATCHER = DESCRIPTOR.message_types_by_name['LabelMatcher']
_QUERYRESULT = DESCRIPTOR.message_types_by_name['QueryResult']
Sample = _reflection.GeneratedProtocolMessageType('Sample', (_message.Message,), {
  'DESCRIPTOR' : _SAMPLE,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Sample)
  })
_sym_db.RegisterMessage(Sample)

LabelPair = _reflection.GeneratedProtocolMessageType('LabelPair', (_message.Message,), {
  'DESCRIPTOR' : _LABELPAIR,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LabelPair)
  })
_sym_db.RegisterMessage(LabelPair)

TimeSeries = _reflection.GeneratedProtocolMessageType('TimeSeries', (_message.Message,), {
  'DESCRIPTOR' : _TIMESERIES,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.TimeSeries)
  })
_sym_db.RegisterMessage(TimeSeries)

WriteRequest = _reflection.GeneratedProtocolMessageType('WriteRequest', (_message.Message,), {
  'DESCRIPTOR' : _WRITEREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.WriteRequest)
  })
_sym_db.RegisterMessage(WriteRequest)

ReadRequest = _reflection.GeneratedProtocolMessageType('ReadRequest', (_message.Message,), {
  'DESCRIPTOR' : _READREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.ReadRequest)
  })
_sym_db.RegisterMessage(ReadRequest)

ReadResponse = _reflection.GeneratedProtocolMessageType('ReadResponse', (_message.Message,), {
  'DESCRIPTOR' : _READRESPONSE,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.ReadResponse)
  })
_sym_db.RegisterMessage(ReadResponse)

Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), {
  'DESCRIPTOR' : _QUERY,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Query)
  })
_sym_db.RegisterMessage(Query)

LabelMatcher = _reflection.GeneratedProtocolMessageType('LabelMatcher', (_message.Message,), {
  'DESCRIPTOR' : _LABELMATCHER,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LabelMatcher)
  })
_sym_db.RegisterMessage(LabelMatcher)

QueryResult = _reflection.GeneratedProtocolMessageType('QueryResult', (_message.Message,), {
  'DESCRIPTOR' : _QUERYRESULT,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.QueryResult)
  })
_sym_db.RegisterMessage(QueryResult)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MATCHTYPE._serialized_start=588
  _MATCHTYPE._serialized_end=662
  _SAMPLE._serialized_start=24
  _SAMPLE._serialized_end=69
  _LABELPAIR._serialized_start=71
  _LABELPAIR._serialized_end=111
  _TIMESERIES._serialized_start=113
  _TIMESERIES._serialized_end=193
  _WRITEREQUEST._serialized_start=195
  _WRITEREQUEST._serialized_end=249
  _READREQUEST._serialized_start=251
  _READREQUEST._serialized_end=296
  _READRESPONSE._serialized_start=298
  _READRESPONSE._serialized_end=350
  _QUERY._serialized_start=352
  _QUERY._serialized_end=453
  _LABELMATCHER._serialized_start=455
  _LABELMATCHER._serialized_end=531
  _QUERYRESULT._serialized_start=533
  _QUERYRESULT._serialized_end=586
# @@protoc_insertion_point(module_scope)
